@@ -42,6 +42,12 @@ class Geometry(ABC):
         Constructs geometric object from combination of Python built-ins.
         """
 
+    @abstractmethod
+    def distance_to(self, other: 'Geometry') -> Coordinate:
+        """
+        Returns distance between geometric objects.
+        """
+
     @abstractmethod
     def raw(self) -> RawGeometry:
         """
