@@ -6,6 +6,7 @@ from clipping.planar import (complete_intersect_multisegments,
                              subtract_multisegments,
                              symmetric_subtract_multisegments,
                              unite_multisegments)
+from locus import segmental
 from orient.planar import (contour_in_contour,
                            multisegment_in_contour,
                            point_in_contour,
@@ -21,32 +22,41 @@ from gon.compound import (Compound,
                           Linear,
                           Location,
                           Relation)
+from gon.core.arithmetic import (non_negative_min,
+                                 robust_divide,
+                                 robust_sqrt)
 from gon.degenerate import EMPTY
-from gon.discrete import (Multipoint,
-                          _robust_divide,
-                          _rotate_points_around_origin,
-                          _rotate_translate_points)
+from gon.discrete import Multipoint
+from gon.discrete.multipoint import (rotate_points_around_origin,
+                                     rotate_translate_points)
 from gon.geometry import Geometry
 from gon.hints import Coordinate
 from gon.primitive import (Point,
-                           RawPoint,
-                           _point_to_step)
+                           RawPoint)
+from gon.primitive.point import point_to_step
 from . import vertices as _vertices
 from .hints import (RawContour,
                     RawMultisegment,
+                    RawSegment,
                     Vertices)
-from .multisegment import Multisegment
+from .multisegment import (Multisegment,
+                           SegmentalSquaredDistanceNode)
+from .raw import (raw_segment_point_distance,
+                  raw_segments_distance,
+                  squared_raw_point_segment_distance,
+                  squared_raw_segments_distance)
 from .segment import Segment
 from .utils import (from_raw_mix_components,
                     from_raw_multisegment,
                     relate_multipoint_to_linear_compound,
-                    robust_sqrt,
                     shift_sequence,
-                    to_pairs_chain)
+                    to_pairs_iterable,
+                    to_pairs_sequence)
 
 
 class Contour(Indexable, Linear):
-    __slots__ = '_raw_locate', '_min_index', '_raw', '_vertices'
+    __slots__ = ('_raw_locate', '_min_index', '_raw', '_vertices',
+                 '_raw_point_nearest_index', '_raw_segment_nearest_index')
 
     def __init__(self, vertices: Vertices) -> None:
         """
@@ -64,6 +74,10 @@ class Contour(Indexable, Linear):
                               key=vertices.__getitem__)
         self._raw = [vertex.raw() for vertex in vertices]
         self._raw_locate = partial(raw_locate_point, self._raw)
+        self._raw_segment_nearest_index = partial(
+                _to_raw_segment_nearest_index, self._raw)
+        self._raw_point_nearest_index = partial(_to_raw_point_nearest_index,
+                                                self._raw)
 
     __repr__ = generate_repr(__init__)
 
@@ -90,7 +104,7 @@ class Contour(Indexable, Linear):
                       if isinstance(other, Multisegment)
                       else
                       (self._intersect_with_raw_multisegment(
-                              to_pairs_chain(other._raw))
+                              to_pairs_sequence(other._raw))
                        if isinstance(other, Contour)
                        else NotImplemented)))
 
@@ -293,7 +307,7 @@ class Contour(Indexable, Linear):
                             if isinstance(other, Multisegment)
                             else
                             (self._unite_with_raw_multisegment(
-                                    to_pairs_chain(other._raw))
+                                    to_pairs_sequence(other._raw))
                              if isinstance(other, Contour)
                              else NotImplemented))))
 
@@ -339,7 +353,7 @@ class Contour(Indexable, Linear):
                             if isinstance(other, Multisegment)
                             else
                             (self._subtract_raw_multisegment(
-                                    to_pairs_chain(other._raw))
+                                    to_pairs_sequence(other._raw))
                              if isinstance(other, Contour)
                              else NotImplemented))))
 
@@ -367,7 +381,7 @@ class Contour(Indexable, Linear):
                        if isinstance(other, Multisegment)
                        else
                        (self._symmetric_subtract_raw_multisegment(
-                               to_pairs_chain(other._raw))
+                               to_pairs_sequence(other._raw))
                         if isinstance(other, Contour)
                         else NotImplemented))))
 
@@ -415,8 +429,8 @@ class Contour(Indexable, Linear):
             accumulated_length += length
             start_x, start_y = end_x, end_y
         divisor = 2 * accumulated_length
-        return Point(_robust_divide(accumulated_x, divisor),
-                     _robust_divide(accumulated_y, divisor))
+        return Point(robust_divide(accumulated_x, divisor),
+                     robust_divide(accumulated_y, divisor))
 
     @property
     def length(self) -> Coordinate:
@@ -470,9 +484,42 @@ class Contour(Indexable, Linear):
         """
         return list(self._vertices)
 
+    def distance_to(self, other: Geometry) -> Coordinate:
+        """
+        Returns distance between the contour and the other geometry.
+
+        Time complexity:
+            ``O(len(self.vertices))``
+        Memory complexity:
+            ``O(1)``
+
+        >>> contour = Contour.from_raw([(0, 0), (1, 0), (0, 1)])
+        >>> contour.distance_to(contour) == 0
+        True
+        """
+        return (self._distance_to_raw_point(other.raw())
+                if isinstance(other, Point)
+                else
+                (non_negative_min(self._distance_to_raw_point(raw_point)
+                                  for raw_point in other._raw)
+                 if isinstance(other, Multipoint)
+                 else
+                 (self._distance_to_raw_segment(other.raw())
+                  if isinstance(other, Segment)
+                  else
+                  (non_negative_min(self._distance_to_raw_segment(raw_segment)
+                                    for raw_segment in other._raw)
+                   if isinstance(other, Multisegment)
+                   else
+                   (non_negative_min(
+                           self._distance_to_raw_segment(raw_segment)
+                           for raw_segment in to_pairs_iterable(other._raw))
+                    if isinstance(other, Contour)
+                    else other.distance_to(self))))))
+
     def index(self) -> None:
         """
-        Pre-processes contour to potentially improve queries.
+        Pre-processes the contour to potentially improve queries.
 
         Time complexity:
             ``O(vertices_count * log vertices_count)`` expected,
@@ -485,8 +532,13 @@ class Contour(Indexable, Linear):
         >>> contour = Contour.from_raw([(0, 0), (1, 0), (0, 1)])
         >>> contour.index()
         """
-        graph = multisegment_trapezoidal(to_pairs_chain(self._raw))
+        raw_contour = self._raw
+        graph = multisegment_trapezoidal(to_pairs_sequence(raw_contour))
         self._raw_locate = graph.locate
+        tree = segmental.Tree(to_pairs_sequence(raw_contour),
+                              node_cls=SegmentalSquaredDistanceNode)
+        self._raw_point_nearest_index = tree.nearest_to_point_index
+        self._raw_segment_nearest_index = tree.nearest_index
 
     def locate(self, point: Point) -> Location:
         """
@@ -588,9 +640,9 @@ class Contour(Indexable, Linear):
         """
         return (rotate_contour_around_origin(self, cosine, sine)
                 if point is None
-                else _rotate_translate_contour(self, cosine, sine,
-                                               *_point_to_step(point, cosine,
-                                                               sine)))
+                else rotate_translate_contour(self, cosine, sine,
+                                              *point_to_step(point, cosine,
+                                                             sine)))
 
     def scale(self,
               factor_x: Coordinate,
@@ -614,9 +666,9 @@ class Contour(Indexable, Linear):
         """
         if factor_y is None:
             factor_y = factor_x
-        return (_scale_contour(self, factor_x, factor_y)
+        return (scale_contour(self, factor_x, factor_y)
                 if factor_x and factor_y
-                else _scale_contour_degenerate(self, factor_x, factor_y))
+                else scale_contour_degenerate(self, factor_x, factor_y))
 
     def to_clockwise(self) -> 'Contour':
         """
@@ -709,29 +761,41 @@ class Contour(Indexable, Linear):
         if edges_intersect(self._raw):
             raise ValueError('Contour should not be self-intersecting.')
 
+    def _distance_to_raw_point(self, other: RawPoint) -> Coordinate:
+        return raw_segment_point_distance(
+                self._raw_edge(self._raw_point_nearest_index(other)), other)
+
+    def _distance_to_raw_segment(self, other: RawSegment) -> Coordinate:
+        return raw_segments_distance(
+                self._raw_edge(self._raw_segment_nearest_index(other)),
+                other)
+
     def _intersect_with_raw_multisegment(self, other_raw: RawMultisegment
                                          ) -> Compound:
         raw_multipoint, raw_multisegment, _ = complete_intersect_multisegments(
-                to_pairs_chain(self._raw), other_raw,
+                to_pairs_sequence(self._raw), other_raw,
                 accurate=False)
         return from_raw_mix_components(raw_multipoint, raw_multisegment)
 
+    def _raw_edge(self, index: int) -> RawSegment:
+        return self._raw[index - 1], self._raw[index]
+
     def _subtract_raw_multisegment(self, other_raw: RawMultisegment
                                    ) -> Compound:
         return from_raw_multisegment(subtract_multisegments(
-                to_pairs_chain(self._raw), other_raw,
+                to_pairs_sequence(self._raw), other_raw,
                 accurate=False))
 
     def _subtract_from_raw_multisegment(self, other_raw: RawMultisegment
                                         ) -> Compound:
         return from_raw_multisegment(subtract_multisegments(
-                other_raw, to_pairs_chain(self._raw),
+                other_raw, to_pairs_sequence(self._raw),
                 accurate=False))
 
     def _symmetric_subtract_raw_multisegment(self, other_raw: RawMultisegment
                                              ) -> Compound:
         return from_raw_multisegment(symmetric_subtract_multisegments(
-                to_pairs_chain(self._raw), other_raw,
+                to_pairs_sequence(self._raw), other_raw,
                 accurate=False))
 
     def _unite_with_multipoint(self, other: Multipoint) -> Compound:
@@ -742,7 +806,7 @@ class Contour(Indexable, Linear):
     def _unite_with_raw_multisegment(self, other_raw: RawMultisegment
                                      ) -> Compound:
         return from_raw_multisegment(unite_multisegments(
-                to_pairs_chain(self._raw), other_raw,
+                to_pairs_sequence(self._raw), other_raw,
                 accurate=False))
 
 
@@ -755,26 +819,62 @@ def raw_locate_point(raw_contour: RawContour, raw_point: RawPoint) -> Location:
 def rotate_contour_around_origin(contour: Contour,
                                  cosine: Coordinate,
                                  sine: Coordinate) -> Contour:
-    return Contour(_rotate_points_around_origin(contour._vertices, cosine,
-                                                sine))
+    return Contour(rotate_points_around_origin(contour._vertices, cosine,
+                                               sine))
 
 
-def _rotate_translate_contour(contour: Contour,
-                              cosine: Coordinate,
-                              sine: Coordinate,
-                              step_x: Coordinate,
-                              step_y: Coordinate) -> Contour:
-    return Contour(_rotate_translate_points(contour._vertices, cosine, sine,
-                                            step_x, step_y))
+def rotate_translate_contour(contour: Contour,
+                             cosine: Coordinate,
+                             sine: Coordinate,
+                             step_x: Coordinate,
+                             step_y: Coordinate) -> Contour:
+    return Contour(rotate_translate_points(contour._vertices, cosine, sine,
+                                           step_x, step_y))
 
 
-def _scale_contour(contour: Contour,
-                   factor_x: Coordinate,
-                   factor_y: Coordinate) -> Contour:
+def scale_contour(contour: Contour,
+                  factor_x: Coordinate,
+                  factor_y: Coordinate) -> Contour:
     return Contour(_vertices.scale(contour._vertices, factor_x, factor_y))
 
 
-def _scale_contour_degenerate(contour: Contour,
-                              factor_x: Coordinate,
-                              factor_y: Coordinate) -> Compound:
+def scale_contour_degenerate(contour: Contour,
+                             factor_x: Coordinate,
+                             factor_y: Coordinate) -> Compound:
     return _vertices.scale_degenerate(contour._vertices, factor_x, factor_y)
+
+
+def _to_raw_point_nearest_index(raw_contour: RawContour,
+                                raw_point: RawPoint) -> int:
+    vertex = raw_contour[-1]
+    enumerated_vertices = enumerate(raw_contour)
+    result, next_vertex = next(enumerated_vertices)
+    squared_distance_to_point = partial(squared_raw_point_segment_distance,
+                                        raw_point)
+    min_squared_distance = squared_distance_to_point((vertex, next_vertex))
+    vertex = next_vertex
+    for index, next_vertex in enumerated_vertices:
+        candidate_squared_distance = squared_distance_to_point((vertex,
+                                                                next_vertex))
+        if candidate_squared_distance < min_squared_distance:
+            result, min_squared_distance = index, candidate_squared_distance
+        vertex = next_vertex
+    return result
+
+
+def _to_raw_segment_nearest_index(raw_contour: RawContour,
+                                  raw_segment: RawSegment) -> int:
+    vertex = raw_contour[-1]
+    enumerated_vertices = enumerate(raw_contour)
+    result, next_vertex = next(enumerated_vertices)
+    squared_distance_to_segment = partial(squared_raw_segments_distance,
+                                          raw_segment)
+    min_squared_distance = squared_distance_to_segment((vertex, next_vertex))
+    vertex = next_vertex
+    for index, next_vertex in enumerated_vertices:
+        candidate_squared_distance = squared_distance_to_segment((vertex,
+                                                                  next_vertex))
+        if candidate_squared_distance < min_squared_distance:
+            result, min_squared_distance = index, candidate_squared_distance
+        vertex = next_vertex
+    return result
