@@ -1,4 +1,5 @@
 from functools import partial
+from math import inf
 from typing import (Iterable,
                     List,
                     Optional)
@@ -9,6 +10,7 @@ from clipping.planar import (complete_intersect_multisegments,
                              subtract_multisegments,
                              symmetric_subtract_multisegments,
                              unite_multisegments)
+from locus import segmental
 from orient.planar import (multisegment_in_multisegment,
                            point_in_multisegment,
                            segment_in_multisegment)
@@ -20,32 +22,39 @@ from gon.compound import (Compound,
                           Linear,
                           Location,
                           Relation)
+from gon.core.arithmetic import (non_negative_min,
+                                 robust_divide,
+                                 robust_sqrt)
+from gon.core.iterable import unique_ever_seen
 from gon.degenerate import EMPTY
-from gon.discrete import (Multipoint,
-                          _robust_divide,
-                          _unique_ever_seen,
-                          from_points)
+from gon.discrete import Multipoint
+from gon.discrete.multipoint import from_points
 from gon.geometry import Geometry
 from gon.hints import Coordinate
 from gon.primitive import (Point,
-                           RawPoint,
-                           _point_to_step,
-                           _scale_point,
-                           _scale_raw_point)
+                           RawPoint)
+from gon.primitive.point import (point_to_step,
+                                 scale_point)
+from gon.primitive.raw import scale_raw_point
 from .hints import (RawMultisegment,
                     RawSegment)
+from .raw import (raw_segment_point_distance,
+                  squared_raw_interval_point_distance,
+                  squared_raw_point_segment_distance,
+                  squared_raw_segment_interval_distance,
+                  squared_raw_segments_distance)
 from .segment import (Segment,
-                      _rotate_translate_segment,
-                      _scale_segment,
-                      rotate_segment_around_origin)
+                      rotate_segment_around_origin,
+                      rotate_translate_segment,
+                      scale_segment)
 from .utils import (from_raw_mix_components,
                     from_raw_multisegment,
-                    relate_multipoint_to_linear_compound,
-                    robust_sqrt)
+                    relate_multipoint_to_linear_compound)
 
 
 class Multisegment(Indexable, Linear):
-    __slots__ = '_segments', '_segments_set', '_raw', '_raw_locate'
+    __slots__ = ('_segments', '_segments_set', '_raw', '_raw_locate',
+                 '_raw_point_nearest_index', '_raw_segment_nearest_index')
 
     def __init__(self, *segments: Segment) -> None:
         """
@@ -62,6 +71,10 @@ class Multisegment(Indexable, Linear):
         self._raw = [segment.raw() for segment in segments]
         self._segments_set = frozenset(segments)
         self._raw_locate = partial(raw_locate_point, self._raw)
+        self._raw_segment_nearest_index = partial(
+                _to_raw_segment_nearest_index, self._raw)
+        self._raw_point_nearest_index = partial(_to_raw_point_nearest_index,
+                                                self._raw)
 
     __repr__ = generate_repr(__init__)
 
@@ -415,8 +428,8 @@ class Multisegment(Indexable, Linear):
             accumulated_y += (start_y + end_y) * length
             accumulated_length += length
         divisor = 2 * accumulated_length
-        return Point(_robust_divide(accumulated_x, divisor),
-                     _robust_divide(accumulated_y, divisor))
+        return Point(robust_divide(accumulated_x, divisor),
+                     robust_divide(accumulated_y, divisor))
 
     @property
     def length(self) -> Coordinate:
@@ -454,9 +467,38 @@ class Multisegment(Indexable, Linear):
         """
         return list(self._segments)
 
+    def distance_to(self, other: Geometry) -> Coordinate:
+        """
+        Returns distance between the multisegment and the other geometry.
+
+        Time complexity:
+            ``O(len(self.segments))``
+        Memory complexity:
+            ``O(1)``
+
+        >>> multisegment = Multisegment.from_raw([((0, 0), (1, 0)),
+        ...                                       ((0, 1), (1, 1))])
+        >>> multisegment.distance_to(multisegment) == 0
+        True
+        """
+        return (self._distance_to_raw_point(other.raw())
+                if isinstance(other, Point)
+                else
+                (non_negative_min(self._distance_to_raw_point(raw_point)
+                                  for raw_point in other._raw)
+                 if isinstance(other, Multipoint)
+                 else
+                 (self._distance_to_raw_segment(other._raw)
+                  if isinstance(other, Segment)
+                  else
+                  (non_negative_min(self._distance_to_raw_segment(raw_segment)
+                                    for raw_segment in other._raw)
+                   if isinstance(other, Multisegment)
+                   else other.distance_to(self)))))
+
     def index(self) -> None:
         """
-        Pre-processes multisegment to potentially improve queries.
+        Pre-processes the multisegment to potentially improve queries.
 
         Time complexity:
             ``O(segments_count * log segments_count)`` expected,
@@ -471,8 +513,13 @@ class Multisegment(Indexable, Linear):
         >>> multisegment.index()
         """
         if len(self._segments) > 1:
-            graph = multisegment_trapezoidal(self._raw)
+            raw_segments = self._raw
+            graph = multisegment_trapezoidal(raw_segments)
             self._raw_locate = graph.locate
+            tree = segmental.Tree(raw_segments,
+                                  node_cls=SegmentalSquaredDistanceNode)
+            self._raw_point_nearest_index = tree.nearest_to_point_index
+            self._raw_segment_nearest_index = tree.nearest_index
 
     def locate(self, point: Point) -> Location:
         """
@@ -565,9 +612,8 @@ class Multisegment(Indexable, Linear):
                                for segment in self._segments])
                 if point is None
                 else Multisegment(
-                *[_rotate_translate_segment(segment, cosine, sine,
-                                            *_point_to_step(point, cosine,
-                                                            sine))
+                *[rotate_translate_segment(segment, cosine, sine,
+                                           *point_to_step(point, cosine, sine))
                   for segment in self._segments]))
 
     def scale(self,
@@ -593,7 +639,7 @@ class Multisegment(Indexable, Linear):
         """
         if factor_y is None:
             factor_y = factor_x
-        return (Multisegment(*[_scale_segment(segment, factor_x, factor_y)
+        return (Multisegment(*[scale_segment(segment, factor_x, factor_y)
                                for segment in self._segments])
                 if factor_x and factor_y
                 else (_scale_segments(self._segments, factor_x, factor_y)
@@ -646,6 +692,14 @@ class Multisegment(Indexable, Linear):
         if segments_cross_or_overlap(self._raw):
             raise ValueError('Crossing or overlapping segments found.')
 
+    def _distance_to_raw_point(self, other: RawPoint) -> Coordinate:
+        return raw_segment_point_distance(
+                self._raw[self._raw_point_nearest_index(other)], other)
+
+    def _distance_to_raw_segment(self, other: RawSegment) -> Coordinate:
+        return robust_sqrt(squared_raw_segments_distance(
+                self._raw[self._raw_segment_nearest_index(other)], other))
+
     def _intersect_with_raw_multisegment(self, other_raw: RawMultisegment
                                          ) -> Compound:
         raw_multipoint, raw_multisegment, _ = complete_intersect_multisegments(
@@ -682,22 +736,23 @@ class Multisegment(Indexable, Linear):
                                                          accurate=False))
 
 
-def _scale_segments(segments: Iterable[Segment],
-                    factor_x: Coordinate,
-                    factor_y: Coordinate) -> Compound:
-    scaled_points, scaled_raw_segments = [], []
-    for segment in segments:
-        if ((factor_x or not segment.is_horizontal) and factor_y
-                or factor_x and not segment.is_vertical):
-            scaled_raw_segments.append((_scale_raw_point(segment.start.raw(),
-                                                         factor_x, factor_y),
-                                        _scale_raw_point(segment.end.raw(),
-                                                         factor_x, factor_y)))
-        else:
-            scaled_points.append(_scale_point(segment.start, factor_x,
-                                              factor_y))
-    return (from_points(_unique_ever_seen(scaled_points))
-            | from_raw_segments(scaled_raw_segments))
+class SegmentalSquaredDistanceNode(segmental.Node):
+    def distance_to_point(self, point: RawPoint,
+                          *,
+                          _minus_inf: Coordinate = -inf) -> Coordinate:
+        return (squared_raw_point_segment_distance(point, self.segment)
+                or _minus_inf
+                if self.is_leaf
+                else squared_raw_interval_point_distance(self.interval, point))
+
+    def distance_to_segment(self, segment: RawSegment,
+                            *,
+                            _minus_inf: Coordinate = -inf) -> Coordinate:
+        return (squared_raw_segments_distance(self.segment, segment)
+                or _minus_inf
+                if self.is_leaf
+                else squared_raw_segment_interval_distance(segment,
+                                                           self.interval))
 
 
 def from_raw_segments(raw_segments: List[RawSegment]) -> Compound:
@@ -713,3 +768,49 @@ def raw_locate_point(raw_multisegment: RawMultisegment,
     return (Location.EXTERIOR
             if relation is Relation.DISJOINT
             else Location.BOUNDARY)
+
+
+def _scale_segments(segments: Iterable[Segment],
+                    factor_x: Coordinate,
+                    factor_y: Coordinate) -> Compound:
+    scaled_points, scaled_raw_segments = [], []
+    for segment in segments:
+        if ((factor_x or not segment.is_horizontal) and factor_y
+                or factor_x and not segment.is_vertical):
+            scaled_raw_segments.append((scale_raw_point(segment.start.raw(),
+                                                        factor_x, factor_y),
+                                        scale_raw_point(segment.end.raw(),
+                                                        factor_x, factor_y)))
+        else:
+            scaled_points.append(scale_point(segment.start, factor_x,
+                                             factor_y))
+    return (from_points(unique_ever_seen(scaled_points))
+            | from_raw_segments(scaled_raw_segments))
+
+
+def _to_raw_point_nearest_index(raw_multisegment: RawMultisegment,
+                                raw_point: RawPoint) -> int:
+    enumerated_candidates = enumerate(raw_multisegment)
+    result, candidate = next(enumerated_candidates)
+    squared_distance_to_point = partial(squared_raw_point_segment_distance,
+                                        raw_point)
+    min_squared_distance = squared_distance_to_point(candidate)
+    for index, candidate in enumerated_candidates:
+        candidate_squared_distance = squared_distance_to_point(candidate)
+        if candidate_squared_distance < min_squared_distance:
+            result, min_squared_distance = index, candidate_squared_distance
+    return result
+
+
+def _to_raw_segment_nearest_index(raw_multisegment: RawMultisegment,
+                                  raw_segment: RawSegment) -> int:
+    enumerated_candidates = enumerate(raw_multisegment)
+    result, candidate = next(enumerated_candidates)
+    squared_distance_to_segment = partial(squared_raw_segments_distance,
+                                          raw_segment)
+    min_squared_distance = squared_distance_to_segment(candidate)
+    for index, candidate in enumerated_candidates:
+        candidate_squared_distance = squared_distance_to_segment(candidate)
+        if candidate_squared_distance < min_squared_distance:
+            result, min_squared_distance = index, candidate_squared_distance
+    return result
