@@ -3,7 +3,6 @@ from typing import Optional
 from orient.planar import (point_in_segment,
                            segment_in_segment)
 from reprit.base import generate_repr
-from robust.hints import Point
 from robust.linear import (segments_intersection,
                            segments_intersections)
 
@@ -11,23 +10,23 @@ from gon.compound import (Compound,
                           Linear,
                           Location,
                           Relation)
+from gon.core.arithmetic import (non_negative_min,
+                                 robust_divide)
 from gon.degenerate import EMPTY
-from gon.discrete import (Multipoint,
-                          _robust_divide)
+from gon.discrete import Multipoint
 from gon.geometry import Geometry
 from gon.hints import Coordinate
-from gon.primitive import (Point,
-                           RawPoint,
-                           _point_to_step,
-                           _rotate_point_around_origin,
-                           _rotate_translate_point,
-                           _scale_point)
+from gon.primitive import Point
+from gon.primitive.point import (point_to_step,
+                                 rotate_point_around_origin,
+                                 rotate_translate_point,
+                                 scale_point)
 from .hints import (RawMultisegment,
                     RawSegment)
+from .raw import (raw_segment_point_distance,
+                  raw_segments_distance)
 from .utils import (from_raw_multisegment,
-                    relate_multipoint_to_linear_compound,
-                    robust_sqrt,
-                    squared_points_distance)
+                    relate_multipoint_to_linear_compound)
 
 
 class Segment(Compound, Linear):
@@ -324,8 +323,8 @@ class Segment(Compound, Linear):
         >>> segment.centroid == Point(1, 0)
         True
         """
-        return Point(_robust_divide(self._start.x + self._end.x, 2),
-                     _robust_divide(self._start.y + self._end.y, 2))
+        return Point(robust_divide(self._start.x + self._end.x, 2),
+                     robust_divide(self._start.y + self._end.y, 2))
 
     @property
     def end(self) -> Point:
@@ -389,7 +388,7 @@ class Segment(Compound, Linear):
         >>> segment.length == 2
         True
         """
-        return robust_sqrt(squared_points_distance(self.start, self.end))
+        return self._end._distance_to_point(self._start)
 
     @property
     def start(self) -> Point:
@@ -407,6 +406,29 @@ class Segment(Compound, Linear):
         """
         return self._start
 
+    def distance_to(self, other: Geometry) -> Coordinate:
+        """
+        Returns distance between the segment and the other geometry.
+
+        Time complexity:
+            ``O(1)``
+        Memory complexity:
+            ``O(1)``
+
+        >>> segment = Segment.from_raw(((0, 0), (2, 0)))
+        >>> segment.distance_to(segment) == 0
+        True
+        """
+        return (raw_segment_point_distance(self._raw, other.raw())
+                if isinstance(other, Point)
+                else (non_negative_min(raw_segment_point_distance(self._raw,
+                                                                  raw_point)
+                                       for raw_point in other._raw)
+                      if isinstance(other, Multipoint)
+                      else (raw_segments_distance(self._raw, other._raw)
+                            if isinstance(other, Segment)
+                            else other.distance_to(self))))
+
     def locate(self, point: Point) -> Location:
         """
         Finds location of the point relative to the segment.
@@ -422,7 +444,9 @@ class Segment(Compound, Linear):
         >>> segment.locate(segment.end) is Location.BOUNDARY
         True
         """
-        return raw_locate_point(self._raw, point.raw())
+        return (Location.BOUNDARY
+                if point_in_segment(point.raw(), self._raw)
+                else Location.EXTERIOR)
 
     def raw(self) -> RawSegment:
         """
@@ -479,8 +503,9 @@ class Segment(Compound, Linear):
         """
         return (rotate_segment_around_origin(self, cosine, sine)
                 if point is None
-                else _rotate_translate_segment(
-                self, cosine, sine, *_point_to_step(point, cosine, sine)))
+                else rotate_translate_segment(self, cosine, sine,
+                                              *point_to_step(point, cosine,
+                                                             sine)))
 
     def scale(self,
               factor_x: Coordinate,
@@ -497,8 +522,8 @@ class Segment(Compound, Linear):
         >>> segment.scale(1) == segment.scale(1, 2) == segment
         True
         """
-        return _scale_segment(self, factor_x,
-                              factor_x if factor_y is None else factor_y)
+        return scale_segment(self, factor_x,
+                             factor_x if factor_y is None else factor_y)
 
     def translate(self, step_x: Coordinate, step_y: Coordinate) -> 'Segment':
         """
@@ -604,37 +629,29 @@ class Segment(Compound, Linear):
 def rotate_segment_around_origin(segment: Segment,
                                  cosine: Coordinate,
                                  sine: Coordinate) -> Segment:
-    return Segment(_rotate_point_around_origin(segment._start, cosine,
-                                               sine),
-                   _rotate_point_around_origin(segment._end, cosine,
-                                               sine))
+    return Segment(rotate_point_around_origin(segment._start, cosine, sine),
+                   rotate_point_around_origin(segment._end, cosine, sine))
 
 
-def _rotate_translate_segment(segment: Segment,
-                              cosine: Coordinate,
-                              sine: Coordinate,
-                              step_x: Coordinate,
-                              step_y: Coordinate) -> Segment:
-    return Segment(_rotate_translate_point(segment._start, cosine, sine,
-                                           step_x, step_y),
-                   _rotate_translate_point(segment._end, cosine, sine,
-                                           step_x, step_y))
+def rotate_translate_segment(segment: Segment,
+                             cosine: Coordinate,
+                             sine: Coordinate,
+                             step_x: Coordinate,
+                             step_y: Coordinate) -> Segment:
+    return Segment(rotate_translate_point(segment._start, cosine, sine, step_x,
+                                          step_y),
+                   rotate_translate_point(segment._end, cosine, sine, step_x,
+                                          step_y))
 
 
-def _scale_segment(segment: Segment,
-                   factor_x: Coordinate,
-                   factor_y: Coordinate) -> Compound:
-    return (Segment(_scale_point(segment._start, factor_x, factor_y),
-                    _scale_point(segment._end, factor_x, factor_y))
+def scale_segment(segment: Segment,
+                  factor_x: Coordinate,
+                  factor_y: Coordinate) -> Compound:
+    return (Segment(scale_point(segment._start, factor_x, factor_y),
+                    scale_point(segment._end, factor_x, factor_y))
             if ((factor_x or not segment.is_horizontal) and factor_y
                 or factor_x and not segment.is_vertical)
-            else Multipoint(_scale_point(segment._start, factor_x, factor_y)))
-
-
-def raw_locate_point(raw_segment: RawSegment, raw_point: RawPoint) -> Location:
-    return (Location.BOUNDARY
-            if point_in_segment(raw_point, raw_segment)
-            else Location.EXTERIOR)
+            else Multipoint(scale_point(segment._start, factor_x, factor_y)))
 
 
 def _raw_subtract_overlap(minuend: RawSegment,
@@ -665,12 +682,6 @@ def _raw_symmetric_subtract_overlap(minuend: RawSegment,
     return [(left_start, left_end), (right_start, right_end)]
 
 
-def _raw_unite_overlap(first_addend: RawSegment,
-                       second_addend: RawSegment) -> RawSegment:
-    start, _, _, end = sorted(first_addend + second_addend)
-    return start, end
-
-
 def _raw_unite_cross(first_addend: RawSegment,
                      second_addend: RawSegment) -> RawMultisegment:
     cross_point = segments_intersection(first_addend, second_addend)
@@ -680,3 +691,9 @@ def _raw_unite_cross(first_addend: RawSegment,
             (second_addend_start, cross_point),
             (cross_point, first_addend_end),
             (cross_point, second_addend_end)]
+
+
+def _raw_unite_overlap(first_addend: RawSegment,
+                       second_addend: RawSegment) -> RawSegment:
+    start, _, _, end = sorted(first_addend + second_addend)
+    return start, end
