@@ -729,9 +729,49 @@ class Mix(Indexable):
         """
         return self._multisegment
 
+    def distance_to(self, other: Geometry) -> Coordinate:
+        """
+        Returns distance between the mix and the other geometry.
+
+        Time complexity:
+            ``O(elements_count)``
+        Memory complexity:
+            ``O(1)``
+
+        where ``elements_count = multisegment_size +\
+ multipolygon_vertices_count``,
+        ``multisegment_size = len(segments)``,
+        ``multipolygon_vertices_count = sum(len(polygon.border.vertices)\
+ + sum(len(hole.vertices) for hole in polygon.holes)\
+ for polygon in polygons)``,
+        ``segments = [] if self.multisegment is EMPTY\
+ else self.multisegment.segments``,
+        ``polygons = [] if self.multipolygon is EMPTY\
+ else self.multipolygon.polygons``.
+
+        >>> mix = Mix.from_raw(([(3, 3), (7, 7)],
+        ...                     [((0, 6), (0, 8)), ((6, 6), (6, 8))],
+        ...                     [([(0, 0), (6, 0), (6, 6), (0, 6)],
+        ...                       [[(2, 2), (2, 4), (4, 4), (4, 2)]])]))
+        >>> mix.distance_to(mix) == 0
+        True
+        """
+        candidates = (component.distance_to(other)
+                      for component in self._components
+                      if component is not EMPTY)
+        result = next(candidates)
+        if not result:
+            return result
+        for candidate in candidates:
+            if not candidate:
+                return candidate
+            elif candidate < result:
+                result = candidate
+        return result
+
     def index(self) -> None:
         """
-        Pre-processes mix to potentially improve queries.
+        Pre-processes the mix to potentially improve queries.
 
         Time complexity:
             ``O(elements_count * log elements_count)`` expected,
@@ -1087,32 +1127,32 @@ class Mix(Indexable):
                 return (Relation.TOUCH
                         if multipoint_relation is Relation.COMPOSITE
                         else multipoint_relation)
-            elif multisegment_relation is Relation.EQUAL:
-                return Relation.COMPONENT
             elif multisegment_relation is Relation.COMPOSITE:
                 multipoint_relation = self._multipoint.relate(other)
                 return (multisegment_relation
                         if multipoint_relation is multisegment_relation
                         else Relation.OVERLAP)
             else:
-                return multisegment_relation
+                return (Relation.COMPONENT
+                        if multisegment_relation is Relation.EQUAL
+                        else multisegment_relation)
         else:
             multipolygon_relation = self._multipolygon.relate(other)
             if multipolygon_relation is Relation.DISJOINT:
                 multisegment_relation = self._multisegment.relate(other)
                 if multisegment_relation is Relation.DISJOINT:
                     multipoint_relation = self._multipoint.relate(other)
-                    return (multipolygon_relation
-                            if multipoint_relation is Relation.DISJOINT
-                            else Relation.TOUCH)
-                elif (multisegment_relation is Relation.TOUCH
-                      or multisegment_relation is Relation.CROSS):
+                    return (Relation.TOUCH
+                            if multipoint_relation is Relation.COMPOSITE
+                            else multipoint_relation)
+                elif multisegment_relation in (Relation.TOUCH,
+                                               Relation.CROSS,
+                                               Relation.COMPONENT):
                     return multisegment_relation
-                elif (multisegment_relation is Relation.EQUAL
-                      or multisegment_relation is Relation.COMPONENT):
-                    return Relation.COMPONENT
                 else:
-                    return Relation.TOUCH
+                    return (Relation.COMPONENT
+                            if multisegment_relation is Relation.EQUAL
+                            else Relation.TOUCH)
             elif (multipolygon_relation is Relation.TOUCH
                   or multipolygon_relation is Relation.CROSS):
                 rest_other = other - self._multipolygon
@@ -1177,20 +1217,14 @@ class Mix(Indexable):
                         else multisegment_relation)
             elif multisegment_relation is Relation.COMPONENT:
                 return (Relation.TOUCH
-                        if (multipoint_relation is Relation.DISJOINT
-                            or multipoint_relation is Relation.TOUCH)
+                        if multipoint_relation is Relation.DISJOINT
                         else (multipoint_relation
-                              if multipoint_relation is Relation.CROSS
+                              if (multipoint_relation is Relation.TOUCH
+                                  or multipoint_relation is Relation.CROSS)
                               else
                               (Relation.COMPOSITE
                                if multipoint_relation is Relation.COMPONENT
                                else Relation.ENCLOSES)))
-            elif multisegment_relation is Relation.ENCLOSED:
-                return (Relation.CROSS
-                        if multipoint_relation in (Relation.DISJOINT,
-                                                   Relation.TOUCH,
-                                                   Relation.CROSS)
-                        else Relation.ENCLOSES)
             else:
                 return (Relation.CROSS
                         if multipoint_relation in (Relation.DISJOINT,
@@ -1211,10 +1245,10 @@ class Mix(Indexable):
                         if other_multipoint_relation in (Relation.DISJOINT,
                                                          Relation.TOUCH,
                                                          Relation.CROSS)
-                        else (Relation.TOUCH
-                              if (other_multipoint_relation
-                                  is Relation.COMPONENT)
-                              else Relation.CROSS))
+                        else
+                        (Relation.TOUCH
+                         if other_multipoint_relation is Relation.COMPONENT
+                         else Relation.CROSS))
             elif other_multisegment_relation is Relation.TOUCH:
                 return (Relation.CROSS
                         if other_multipoint_relation in (Relation.CROSS,
@@ -1241,9 +1275,10 @@ class Mix(Indexable):
                         if other_multipoint_relation in (Relation.DISJOINT,
                                                          Relation.TOUCH,
                                                          Relation.CROSS)
-                        else (other_multisegment_relation
-                              if other_multipoint_relation is Relation.WITHIN
-                              else Relation.ENCLOSED))
+                        else
+                        (Relation.ENCLOSED
+                         if other_multipoint_relation is Relation.COMPONENT
+                         else other_multisegment_relation))
         multipolygons_relation = self._multipolygon.relate(
                 other._multipolygon)
         if (multipolygons_relation is Relation.DISJOINT
@@ -1635,16 +1670,6 @@ class Mix(Indexable):
                         return (Relation.COMPONENT
                                 if multipolygon_relation is Relation.TOUCH
                                 else Relation.ENCLOSED)
-            elif self._multipoint is EMPTY:
-                return self._multisegment.relate(other)
-            elif self._multisegment is EMPTY:
-                multipoint_relation = self._multipoint.relate(other)
-                return (multipolygon_relation
-                        if multipoint_relation is Relation.DISJOINT
-                        else (Relation.COMPONENT
-                              if (multipoint_relation is Relation.EQUAL
-                                  or multipoint_relation is Relation.COMPONENT)
-                              else Relation.TOUCH))
             else:
                 multisegment_relation = self._multisegment.relate(other)
                 if multisegment_relation is Relation.DISJOINT:
@@ -1682,24 +1707,16 @@ class Mix(Indexable):
                                     if (multipoint_relation
                                         is Relation.COMPOSITE)
                                     else Relation.CROSS)))
-            elif multisegment_relation is Relation.COVER:
+            elif (multisegment_relation is Relation.COVER
+                  or multisegment_relation is Relation.ENCLOSES):
                 multipoint_relation = self._multipoint.relate(other)
-                return (multipoint_relation
-                        if (multipoint_relation is Relation.CROSS
-                            or multipoint_relation is Relation.COVER)
-                        else (Relation.ENCLOSES
-                              if (multipoint_relation is Relation.ENCLOSES
-                                  or multipoint_relation is Relation.COMPOSITE)
-                              else Relation.CROSS))
-            elif multisegment_relation is Relation.ENCLOSES:
-                multipoint_relation = self._multipoint.relate(other)
-                return (multipoint_relation
-                        if (multipoint_relation is Relation.CROSS
-                            or multipoint_relation is Relation.ENCLOSES)
-                        else (multisegment_relation
-                              if (multipoint_relation is Relation.COVER
-                                  or multipoint_relation is Relation.COMPOSITE)
-                              else Relation.CROSS))
+                return (Relation.CROSS
+                        if (multipoint_relation is Relation.DISJOINT
+                            or multipoint_relation is Relation.TOUCH)
+                        else (multipoint_relation
+                              if (multipoint_relation is multisegment_relation
+                                  or multipoint_relation is Relation.CROSS)
+                              else Relation.ENCLOSES))
             elif multisegment_relation is Relation.COMPOSITE:
                 multipoint_relation = self._multipoint.relate(other)
                 return (Relation.TOUCH
@@ -1709,56 +1726,107 @@ class Mix(Indexable):
                                   or multipoint_relation is Relation.CROSS)
                               else (multisegment_relation
                                     if (multipoint_relation
-                                        is Relation.COMPOSITE)
+                                        is multisegment_relation)
                                     else Relation.CROSS)))
             else:
                 return multisegment_relation
         else:
             multipolygon_relation = self._multipolygon.relate(other)
-            if multipolygon_relation in (Relation.OVERLAP,
-                                         Relation.COMPONENT,
-                                         Relation.ENCLOSED,
-                                         Relation.WITHIN):
-                return multipolygon_relation
-            elif multipolygon_relation is Relation.EQUAL:
-                return Relation.COMPONENT
-            elif self._multisegment is EMPTY:
-                multipoint_relation = self._multipoint.relate(other)
-                return (multipolygon_relation
-                        if multipoint_relation is multipolygon_relation
-                        else (Relation.ENCLOSES
-                              if multipoint_relation in (Relation.COVER,
-                                                         Relation.ENCLOSES,
-                                                         Relation.COMPOSITE)
-                              else Relation.OVERLAP))
-            else:
+            if multipolygon_relation is Relation.DISJOINT:
+                multisegment_relation = self._multisegment.relate(other)
+                if multisegment_relation is Relation.DISJOINT:
+                    multipoint_relation = self._multipoint.relate(other)
+                    return (multipoint_relation
+                            if multipoint_relation in (Relation.DISJOINT,
+                                                       Relation.TOUCH,
+                                                       Relation.CROSS)
+                            else (Relation.TOUCH
+                                  if multipoint_relation is Relation.COMPOSITE
+                                  else Relation.CROSS))
+                elif (multisegment_relation is Relation.TOUCH
+                      or multisegment_relation is Relation.COMPOSITE):
+                    multipoint_relation = self._multipoint.relate(other)
+                    return (Relation.TOUCH
+                            if multipoint_relation in (Relation.DISJOINT,
+                                                       Relation.TOUCH,
+                                                       Relation.COMPOSITE)
+                            else Relation.CROSS)
+                else:
+                    return Relation.CROSS
+            elif multipolygon_relation is Relation.TOUCH:
                 multisegment_relation = self._multisegment.relate(other)
                 if multisegment_relation in (Relation.DISJOINT,
                                              Relation.TOUCH,
-                                             Relation.CROSS):
-                    return Relation.OVERLAP
-                elif multisegment_relation is multipolygon_relation:
-                    if self._multipoint is EMPTY:
+                                             Relation.COMPOSITE):
+                    multipoint_relation = self._multipoint.relate(other)
+                    return (multipolygon_relation
+                            if multipoint_relation in (Relation.DISJOINT,
+                                                       Relation.TOUCH,
+                                                       Relation.COMPOSITE)
+                            else Relation.CROSS)
+                else:
+                    return Relation.CROSS
+            elif (multipolygon_relation is Relation.COVER
+                  or multipolygon_relation is Relation.COMPOSITE):
+                if self._multisegment is EMPTY:
+                    multipoint_relation = self._multipoint.relate(other)
+                    return (Relation.OVERLAP
+                            if multipoint_relation in (Relation.DISJOINT,
+                                                       Relation.TOUCH,
+                                                       Relation.CROSS)
+                            else (multipolygon_relation
+                                  if (multipoint_relation
+                                      is multipolygon_relation)
+                                  else Relation.ENCLOSES))
+                else:
+                    multisegment_relation = self._multisegment.relate(other)
+                    if multisegment_relation in (Relation.DISJOINT,
+                                                 Relation.TOUCH,
+                                                 Relation.CROSS):
+                        return Relation.OVERLAP
+                    elif self._multipoint is EMPTY:
+                        return (multipolygon_relation
+                                if (multisegment_relation
+                                    is multipolygon_relation)
+                                else Relation.ENCLOSES)
+                    else:
+                        multipoint_relation = self._multipoint.relate(other)
+                        return (Relation.OVERLAP
+                                if multipoint_relation in (Relation.DISJOINT,
+                                                           Relation.TOUCH,
+                                                           Relation.CROSS)
+                                else (multipolygon_relation
+                                      if (multipoint_relation
+                                          is multisegment_relation
+                                          is multipolygon_relation)
+                                      else Relation.ENCLOSES))
+            elif multipolygon_relation is Relation.ENCLOSES:
+                if self._multisegment is EMPTY:
+                    multipoint_relation = self._multipoint.relate(other)
+                    return (Relation.OVERLAP
+                            if multipoint_relation in (Relation.DISJOINT,
+                                                       Relation.TOUCH,
+                                                       Relation.CROSS)
+                            else Relation.ENCLOSES)
+                else:
+                    multisegment_relation = self._multisegment.relate(other)
+                    if multisegment_relation in (Relation.DISJOINT,
+                                                 Relation.TOUCH,
+                                                 Relation.CROSS):
+                        return Relation.OVERLAP
+                    elif self._multipoint is EMPTY:
                         return multipolygon_relation
                     else:
                         multipoint_relation = self._multipoint.relate(other)
-                        return (multipolygon_relation
-                                if multipoint_relation is multipolygon_relation
-                                else
-                                (Relation.ENCLOSES
-                                 if multipoint_relation in (Relation.COVER,
-                                                            Relation.ENCLOSES,
-                                                            Relation.COMPOSITE)
-                                 else Relation.OVERLAP))
-                elif self._multipoint is EMPTY:
-                    return Relation.ENCLOSES
-                else:
-                    multipoint_relation = self._multipoint.relate(other)
-                    return (Relation.ENCLOSES
-                            if multipoint_relation in (Relation.COVER,
-                                                       Relation.ENCLOSES,
-                                                       Relation.COMPOSITE)
-                            else Relation.OVERLAP)
+                        return (Relation.OVERLAP
+                                if multipoint_relation in (Relation.DISJOINT,
+                                                           Relation.TOUCH,
+                                                           Relation.CROSS)
+                                else Relation.ENCLOSES)
+            else:
+                return (Relation.COMPONENT
+                        if multipolygon_relation is Relation.EQUAL
+                        else multipolygon_relation)
 
 
 def from_mix_components(multipoint: Maybe[Multipoint],
