@@ -6,6 +6,7 @@ from typing import (List,
 
 from bentley_ottmann.planar import segments_cross_or_overlap
 from clipping.planar import (complete_intersect_multipolygons,
+                             complete_intersect_multiregions,
                              complete_intersect_multisegment_with_multipolygon,
                              subtract_multipolygon_from_multisegment,
                              subtract_multipolygons,
@@ -26,27 +27,34 @@ from gon.compound import (Compound,
                           Linear,
                           Relation,
                           Shaped)
+from gon.core.arithmetic import (non_negative_min,
+                                 robust_divide)
 from gon.degenerate import EMPTY
-from gon.discrete import (Multipoint,
-                          _robust_divide)
+from gon.discrete import Multipoint
 from gon.geometry import Geometry
 from gon.hints import Coordinate
 from gon.linear import (Contour,
                         Multisegment,
                         RawMultisegment,
+                        RawSegment,
                         Segment,
                         vertices)
 from gon.linear.utils import (from_raw_multisegment,
-                              to_pairs_chain)
+                              to_pairs_iterable,
+                              to_pairs_sequence)
 from gon.primitive import (Point,
-                           _point_to_step)
-from .hints import RawMultipolygon
+                           RawPoint)
+from gon.primitive.point import point_to_step
+from .hints import (RawMultipolygon,
+                    RawMultiregion)
 from .polygon import (Polygon,
-                      _polygon_to_centroid_components,
-                      _rotate_translate_polygon,
-                      _scale_polygon,
-                      rotate_polygon_around_origin)
+                      polygon_to_centroid_components,
+                      polygon_to_raw_edges,
+                      rotate_polygon_around_origin,
+                      rotate_translate_polygon,
+                      scale_polygon)
 from .utils import (flatten,
+                    from_raw_holeless_mix_components,
                     from_raw_mix_components,
                     from_raw_multipolygon)
 
@@ -70,7 +78,7 @@ class Multipolygon(Indexable, Shaped):
         self._polygons = polygons
         self._polygons_set = frozenset(polygons)
         self._raw = [polygon.raw() for polygon in polygons]
-        self._locate = partial(locate_point_in_polygons, self._polygons)
+        self._locate = partial(_locate_point_in_polygons, self._polygons)
 
     __repr__ = generate_repr(__init__)
 
@@ -101,7 +109,7 @@ class Multipolygon(Indexable, Shaped):
                       if isinstance(other, Multisegment)
                       else
                       (self._intersect_with_raw_multisegment(
-                              to_pairs_chain(other.raw()))
+                              to_pairs_sequence(other.raw()))
                        if isinstance(other, Contour)
                        else
                        (self._intersect_with_raw_multipolygon([other.raw()])
@@ -325,7 +333,7 @@ class Multipolygon(Indexable, Shaped):
                   if isinstance(other, Multisegment)
                   else
                   (self._unite_with_raw_multisegment(
-                          to_pairs_chain(other.raw()))
+                          to_pairs_sequence(other.raw()))
                    if isinstance(other, Contour)
                    else (self._unite_with_raw_multipolygon([other.raw()])
                          if isinstance(other, Polygon)
@@ -354,7 +362,7 @@ class Multipolygon(Indexable, Shaped):
                       if isinstance(other, Multisegment)
                       else
                       (self._subtract_from_raw_multisegment(
-                              to_pairs_chain(other.raw()))
+                              to_pairs_sequence(other.raw()))
                        if isinstance(other, Contour)
                        else (self._subtract_from_raw_multipolygon(
                               [other.raw()])
@@ -418,7 +426,7 @@ class Multipolygon(Indexable, Shaped):
                   if isinstance(other, Multisegment)
                   else
                   (self._unite_with_raw_multisegment(
-                          to_pairs_chain(other.raw()))
+                          to_pairs_sequence(other.raw()))
                    if isinstance(other, Contour)
                    else
                    (self._symmetric_subtract_raw_multipolygon([other.raw()])
@@ -499,17 +507,17 @@ class Multipolygon(Indexable, Shaped):
         """
         polygons = iter(self._polygons)
         (x_numerator, y_numerator,
-         double_area) = _polygon_to_centroid_components(next(polygons))
+         double_area) = polygon_to_centroid_components(next(polygons))
         for polygon in polygons:
             (polygon_x_numerator, polygon_y_numerator,
-             polygon_double_area) = _polygon_to_centroid_components(polygon)
+             polygon_double_area) = polygon_to_centroid_components(polygon)
             x_numerator, y_numerator, double_area = (
                 sum_expansions(x_numerator, polygon_x_numerator),
                 sum_expansions(y_numerator, polygon_y_numerator),
                 sum_expansions(double_area, polygon_double_area))
         divisor = 3 * double_area[-1]
-        return Point(_robust_divide(x_numerator[-1], divisor),
-                     _robust_divide(y_numerator[-1], divisor))
+        return Point(robust_divide(x_numerator[-1], divisor),
+                     robust_divide(y_numerator[-1], divisor))
 
     @property
     def perimeter(self) -> Coordinate:
@@ -555,9 +563,61 @@ class Multipolygon(Indexable, Shaped):
         """
         return list(self._polygons)
 
+    def distance_to(self, other: Geometry) -> Coordinate:
+        """
+        Returns distance between the multipolygon and the other geometry.
+
+        Time complexity:
+            ``O(vertices_count)``
+        Memory complexity:
+            ``O(1)``
+
+        where ``vertices_count = sum(len(polygon.border.vertices)\
+ + sum(len(hole.vertices) for hole in polygon.holes)\
+ for polygon in self.polygons)``.
+
+        >>> multipolygon = Multipolygon.from_raw(
+        ...         [([(0, 0), (6, 0), (6, 6), (0, 6)],
+        ...           [[(2, 2), (2, 4), (4, 4), (4, 2)]])])
+        >>> multipolygon.distance_to(multipolygon) == 0
+        True
+        """
+        return (self._distance_to_raw_point(other.raw())
+                if isinstance(other, Point)
+                else
+                (non_negative_min(self._distance_to_raw_point(raw_point)
+                                  for raw_point in other._raw)
+                 if isinstance(other, Multipoint)
+                 else
+                 (self._distance_to_raw_segment(other.raw())
+                  if isinstance(other, Segment)
+                  else
+                  (non_negative_min(self._distance_to_raw_segment(raw_segment)
+                                    for raw_segment in other._raw)
+                   if isinstance(other, Multisegment)
+                   else
+                   (non_negative_min(
+                           self._distance_to_raw_segment(raw_segment)
+                           for raw_segment in to_pairs_iterable(other._raw))
+                    if isinstance(other, Contour)
+                    else
+                    (non_negative_min(
+                            self._distance_to_raw_segment(raw_segment)
+                            for raw_segment in polygon_to_raw_edges(other))
+                     if isinstance(other, Polygon)
+                     else
+                     ((non_negative_min(
+                             self._distance_to_raw_segment(raw_segment)
+                             for polygon in other._polygons
+                             for raw_segment in polygon_to_raw_edges(polygon))
+                       if self.disjoint(other)
+                       else 0)
+                      if isinstance(other, Multipolygon)
+                      else other.distance_to(self))))))))
+
     def index(self) -> None:
         """
-        Pre-processes multipolygon to potentially improve queries.
+        Pre-processes the multipolygon to potentially improve queries.
 
         Time complexity:
             ``O(vertices_count * log vertices_count)`` expected,
@@ -579,18 +639,24 @@ class Multipolygon(Indexable, Shaped):
             polygon.index()
 
         def polygon_to_interval(polygon: Polygon) -> r.Interval:
-            vertices = iter(polygon.border.vertices)
-            first_vertex = next(vertices)
+            border_vertices = iter(polygon.border._vertices)
+            first_vertex = next(border_vertices)
             x_min = x_max = first_vertex.x
             y_min = y_max = first_vertex.y
-            for vertex in vertices:
-                x_min, x_max = min(x_min, vertex.x), max(x_max, vertex.x)
-                y_min, y_max = min(y_min, vertex.y), max(y_max, vertex.y)
+            for vertex in border_vertices:
+                if vertex.x < x_min:
+                    x_min = vertex.x
+                elif vertex.x > x_max:
+                    x_max = vertex.x
+                if vertex.y < y_min:
+                    y_min = vertex.y
+                elif vertex.y > y_max:
+                    y_max = vertex.y
             return (x_min, x_max), (y_min, y_max)
 
         tree = r.Tree([polygon_to_interval(polygon) for polygon in polygons])
-        self._locate = partial(locate_point_in_indexed_polygons,
-                               tree, polygons)
+        self._locate = partial(_locate_point_in_indexed_polygons, tree,
+                               polygons)
 
     def locate(self, point: Point) -> Location:
         """
@@ -711,10 +777,10 @@ class Multipolygon(Indexable, Shaped):
         ...           [[(0, 2), (-2, 2), (-2, 4), (0, 4)]])])
         True
         """
-        return (rotate_multipolygon_around_origin(self, cosine, sine)
+        return (_rotate_multipolygon_around_origin(self, cosine, sine)
                 if point is None
                 else _rotate_translate_multipolygon(
-                self, cosine, sine, *_point_to_step(point, cosine, sine)))
+                self, cosine, sine, *point_to_step(point, cosine, sine)))
 
     def scale(self,
               factor_x: Coordinate,
@@ -743,7 +809,7 @@ class Multipolygon(Indexable, Shaped):
         """
         if factor_y is None:
             factor_y = factor_x
-        return (Multipolygon(*[_scale_polygon(polygon, factor_x, factor_y)
+        return (Multipolygon(*[scale_polygon(polygon, factor_x, factor_y)
                                for polygon in self._polygons])
                 if factor_x and factor_y
                 else vertices.scale_degenerate(
@@ -832,8 +898,8 @@ class Multipolygon(Indexable, Shaped):
         for polygon in self._polygons:
             polygon.validate()
         polygons_raw_edges = list(flatten(
-                chain(to_pairs_chain(polygon.border.raw()),
-                      flatten(to_pairs_chain(hole.raw())
+                chain(to_pairs_sequence(polygon.border.raw()),
+                      flatten(to_pairs_sequence(hole.raw())
                               for hole in polygon.holes))
                 for polygon in self._polygons))
         if segments_cross_or_overlap(polygons_raw_edges,
@@ -841,11 +907,27 @@ class Multipolygon(Indexable, Shaped):
             raise ValueError('Polygons should only touch each other '
                              'in discrete number of points.')
 
+    def _distance_to_raw_point(self, other: RawPoint) -> Coordinate:
+        return non_negative_min(polygon._distance_to_raw_point(other)
+                                for polygon in self._polygons)
+
+    def _distance_to_raw_segment(self, other: RawSegment) -> Coordinate:
+        return non_negative_min(polygon._distance_to_raw_segment(other)
+                                for polygon in self._polygons)
+
     def _intersect_with_raw_multipolygon(self, other_raw: RawMultipolygon
                                          ) -> Compound:
-        return from_raw_mix_components(
-                *complete_intersect_multipolygons(self._raw, other_raw,
+        raw = self._raw
+        return (from_raw_mix_components(
+                *complete_intersect_multipolygons(raw, other_raw,
                                                   accurate=False))
+                if (_raw_multipolygon_has_holes(raw)
+                    or _raw_multipolygon_has_holes(other_raw))
+                else from_raw_holeless_mix_components(
+                *complete_intersect_multiregions(
+                        _raw_multipolygon_to_multiregion(raw),
+                        _raw_multipolygon_to_multiregion(other_raw),
+                        accurate=False)))
 
     def _intersect_with_raw_multisegment(self,
                                          raw_multisegment: RawMultisegment
@@ -897,8 +979,16 @@ class Multipolygon(Indexable, Shaped):
                                                          accurate=False))
 
 
-def locate_point_in_polygons(polygons: Sequence[Polygon],
-                             point: Point) -> Location:
+def _raw_multipolygon_has_holes(raw: RawMultipolygon) -> bool:
+    return any(holes for _, holes in raw)
+
+
+def _raw_multipolygon_to_multiregion(raw: RawMultipolygon) -> RawMultiregion:
+    return [border for border, _ in raw]
+
+
+def _locate_point_in_polygons(polygons: Sequence[Polygon],
+                              point: Point) -> Location:
     for polygon in polygons:
         location = polygon.locate(point)
         if location is not Location.EXTERIOR:
@@ -906,8 +996,9 @@ def locate_point_in_polygons(polygons: Sequence[Polygon],
     return Location.EXTERIOR
 
 
-def locate_point_in_indexed_polygons(tree: r.Tree, polygons: Sequence[Polygon],
-                                     point: Point) -> Location:
+def _locate_point_in_indexed_polygons(tree: r.Tree,
+                                      polygons: Sequence[Polygon],
+                                      point: Point) -> Location:
     candidates_indices = tree.find_supersets_indices(((point.x, point.x),
                                                       (point.y, point.y)))
     for candidate_index in candidates_indices:
@@ -917,9 +1008,9 @@ def locate_point_in_indexed_polygons(tree: r.Tree, polygons: Sequence[Polygon],
     return Location.EXTERIOR
 
 
-def rotate_multipolygon_around_origin(multipolygon: Multipolygon,
-                                      cosine: Coordinate,
-                                      sine: Coordinate) -> Multipolygon:
+def _rotate_multipolygon_around_origin(multipolygon: Multipolygon,
+                                       cosine: Coordinate,
+                                       sine: Coordinate) -> Multipolygon:
     return Multipolygon(*[rotate_polygon_around_origin(polygon, cosine, sine)
                           for polygon in multipolygon._polygons])
 
@@ -929,6 +1020,6 @@ def _rotate_translate_multipolygon(multipolygon: Multipolygon,
                                    sine: Coordinate,
                                    step_x: Coordinate,
                                    step_y: Coordinate) -> Multipolygon:
-    return Multipolygon(*[_rotate_translate_polygon(polygon, cosine, sine,
-                                                    step_x, step_y)
+    return Multipolygon(*[rotate_translate_polygon(polygon, cosine, sine,
+                                                   step_x, step_y)
                           for polygon in multipolygon._polygons])
