@@ -1,15 +1,21 @@
+from bisect import bisect
 from functools import partial
-from typing import (List,
+from itertools import (accumulate,
+                       chain)
+from typing import (Iterator,
+                    List,
                     Optional,
                     Sequence,
                     Tuple)
 
 from clipping.planar import (complete_intersect_multipolygons,
+                             complete_intersect_multiregions,
                              complete_intersect_multisegment_with_multipolygon,
                              subtract_multipolygon_from_multisegment,
                              subtract_multipolygons,
                              symmetric_subtract_multipolygons,
                              unite_multipolygons)
+from locus import segmental
 from orient.planar import (contour_in_polygon,
                            multisegment_in_polygon,
                            point_in_polygon,
@@ -31,36 +37,47 @@ from gon.compound import (Compound,
                           Location,
                           Relation,
                           Shaped)
+from gon.core.arithmetic import (non_negative_min,
+                                 robust_divide)
 from gon.degenerate import EMPTY
-from gon.discrete import (Multipoint,
-                          _robust_divide)
+from gon.discrete import Multipoint
 from gon.geometry import Geometry
 from gon.hints import Coordinate
 from gon.linear import (Contour,
                         Multisegment,
                         RawContour,
                         RawMultisegment,
+                        RawSegment,
                         Segment,
                         vertices)
-from gon.linear.contour import (_rotate_translate_contour,
-                                _scale_contour,
-                                _scale_contour_degenerate,
-                                rotate_contour_around_origin)
+from gon.linear.contour import (rotate_contour_around_origin,
+                                rotate_translate_contour,
+                                scale_contour,
+                                scale_contour_degenerate)
+from gon.linear.multisegment import SegmentalSquaredDistanceNode
+from gon.linear.raw import (raw_segment_point_distance,
+                            raw_segments_distance,
+                            squared_raw_point_segment_distance,
+                            squared_raw_segments_distance)
 from gon.linear.utils import (from_raw_multisegment,
-                              to_pairs_chain)
+                              to_pairs_iterable,
+                              to_pairs_sequence)
 from gon.primitive import (Point,
-                           RawPoint,
-                           _point_to_step)
+                           RawPoint)
+from gon.primitive.point import point_to_step
 from .hints import (RawMultipolygon,
                     RawPolygon)
-from .utils import (from_raw_mix_components,
+from .utils import (flatten,
+                    from_raw_holeless_mix_components,
+                    from_raw_mix_components,
                     from_raw_multipolygon,
                     to_convex_hull)
 
 
 class Polygon(Indexable, Shaped):
-    __slots__ = ('_border', '_holes', '_holes_set',
-                 '_raw_border', '_raw_holes', '_raw_locate')
+    __slots__ = ('_border', '_holes', '_holes_set', '_raw_border',
+                 '_raw_holes', '_raw_locate', '_raw_point_nearest_path',
+                 '_raw_segment_nearest_path')
 
     def __init__(self, border: Contour,
                  holes: Optional[Sequence[Contour]] = None) -> None:
@@ -82,6 +99,12 @@ class Polygon(Indexable, Shaped):
                                                            for hole in holes]
         self._raw_locate = partial(raw_locate_point,
                                    (self._raw_border, self._raw_holes))
+        self._raw_segment_nearest_path = partial(_to_raw_segment_nearest_path,
+                                                 self._raw_border,
+                                                 self._raw_holes)
+        self._raw_point_nearest_path = partial(_to_raw_point_nearest_path,
+                                               self._raw_border,
+                                               self._raw_holes)
 
     __repr__ = generate_repr(__init__)
 
@@ -108,12 +131,21 @@ class Polygon(Indexable, Shaped):
                       if isinstance(other, Multisegment)
                       else
                       (self._intersect_with_raw_multisegment(
-                              to_pairs_chain(other.raw()))
+                              to_pairs_sequence(other.raw()))
                        if isinstance(other, Contour)
-                       else (self._intersect_with_raw_multipolygon(
-                              [(other._raw_border, other._raw_holes)])
-                             if isinstance(other, Polygon)
-                             else NotImplemented))))
+                       else
+                       ((from_raw_mix_components(
+                               *complete_intersect_multipolygons(
+                                       [(self._raw_border, self._raw_holes)],
+                                       [(other._raw_border, other._raw_holes)],
+                                       accurate=False))
+                         if self._holes or other._holes
+                         else from_raw_holeless_mix_components(
+                               *complete_intersect_multiregions(
+                                       [self._raw_border], [other._raw_border],
+                                       accurate=False)))
+                        if isinstance(other, Polygon)
+                        else NotImplemented))))
 
     __rand__ = __and__
 
@@ -321,7 +353,7 @@ class Polygon(Indexable, Shaped):
                   if isinstance(other, Multisegment)
                   else
                   (self._unite_with_raw_multisegment(
-                          to_pairs_chain(other.raw()))
+                          to_pairs_sequence(other.raw()))
                    if isinstance(other, Contour)
                    else (self._unite_with_raw_multipolygon([other.raw()])
                          if isinstance(other, Polygon)
@@ -347,7 +379,7 @@ class Polygon(Indexable, Shaped):
                       if isinstance(other, Multisegment)
                       else
                       (self._subtract_from_raw_multisegment(
-                              to_pairs_chain(other.raw()))
+                              to_pairs_sequence(other.raw()))
                        if isinstance(other, Contour)
                        else NotImplemented)))
 
@@ -402,7 +434,7 @@ class Polygon(Indexable, Shaped):
                   if isinstance(other, Multisegment)
                   else
                   (self._unite_with_raw_multisegment(
-                          to_pairs_chain(other.raw()))
+                          to_pairs_sequence(other.raw()))
                    if isinstance(other, Contour)
                    else
                    (self._symmetric_subtract_raw_multipolygon([other.raw()])
@@ -495,11 +527,11 @@ class Polygon(Indexable, Shaped):
         >>> polygon.centroid == Point(3, 3)
         True
         """
-        x_numerator, y_numerator, double_area = _polygon_to_centroid_components(
+        x_numerator, y_numerator, double_area = polygon_to_centroid_components(
                 self)
         divisor = 3 * double_area[-1]
-        return Point(_robust_divide(x_numerator[-1], divisor),
-                     _robust_divide(y_numerator[-1], divisor))
+        return Point(robust_divide(x_numerator[-1], divisor),
+                     robust_divide(y_numerator[-1], divisor))
 
     @property
     def convex_hull(self) -> 'Polygon':
@@ -584,9 +616,53 @@ class Polygon(Indexable, Shaped):
         """
         return self._border.length + sum(hole.length for hole in self._holes)
 
+    def distance_to(self, other: Geometry) -> Coordinate:
+        """
+        Returns distance between the polygon and the other geometry.
+
+        Time complexity:
+            ``O(vertices_count)``
+        Memory complexity:
+            ``O(1)``
+
+        where ``vertices_count = len(self.border.vertices)\
+ + sum(len(hole.vertices) for hole in self.holes)``.
+
+        >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
+        ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
+        >>> polygon.distance_to(polygon) == 0
+        True
+        """
+        return (self._distance_to_raw_point(other.raw())
+                if isinstance(other, Point)
+                else
+                (non_negative_min(self._distance_to_raw_point(raw_point)
+                                  for raw_point in other._raw)
+                 if isinstance(other, Multipoint)
+                 else
+                 (self._distance_to_raw_segment(other.raw())
+                  if isinstance(other, Segment)
+                  else
+                  (non_negative_min(self._distance_to_raw_segment(raw_segment)
+                                    for raw_segment in other._raw)
+                   if isinstance(other, Multisegment)
+                   else
+                   (non_negative_min(
+                           self._distance_to_raw_segment(raw_segment)
+                           for raw_segment in to_pairs_iterable(other._raw))
+                    if isinstance(other, Contour)
+                    else
+                    ((non_negative_min(
+                            self._linear_distance_to_raw_segment(raw_segment)
+                            for raw_segment in polygon_to_raw_edges(other))
+                      if self.disjoint(other)
+                      else 0)
+                     if isinstance(other, Polygon)
+                     else other.distance_to(self)))))))
+
     def index(self) -> None:
         """
-        Pre-processes polygon to potentially improve queries.
+        Pre-processes the polygon to potentially improve queries.
 
         Time complexity:
             ``O(vertices_count * log vertices_count)`` expected,
@@ -603,6 +679,21 @@ class Polygon(Indexable, Shaped):
         """
         graph = polygon_trapezoidal(self._raw_border, self._raw_holes)
         self._raw_locate = graph.locate
+        tree = segmental.Tree(list(polygon_to_raw_edges(self)),
+                              node_cls=SegmentalSquaredDistanceNode)
+        if self._raw_holes:
+            contours_offsets = tuple(
+                    accumulate(chain((0, len(self._raw_border)),
+                                     map(len, self._raw_holes))))
+            self._raw_point_nearest_path, self._raw_segment_nearest_path = (
+                partial(_tree_to_raw_point_nearest_path, tree,
+                        contours_offsets),
+                partial(_tree_to_raw_segment_nearest_path, tree,
+                        contours_offsets))
+        else:
+            self._raw_point_nearest_path, self._raw_segment_nearest_path = (
+                partial(_tree_to_holeless_raw_point_nearest_path, tree),
+                partial(_tree_to_holeless_raw_segment_nearest_path, tree))
 
     def locate(self, point: Point) -> Location:
         """
@@ -714,9 +805,9 @@ class Polygon(Indexable, Shaped):
         """
         return (rotate_polygon_around_origin(self, cosine, sine)
                 if point is None
-                else _rotate_translate_polygon(self, cosine, sine,
-                                               *_point_to_step(point, cosine,
-                                                               sine)))
+                else rotate_translate_polygon(self, cosine, sine,
+                                              *point_to_step(point, cosine,
+                                                             sine)))
 
     def scale(self,
               factor_x: Coordinate,
@@ -743,10 +834,10 @@ class Polygon(Indexable, Shaped):
         """
         if factor_y is None:
             factor_y = factor_x
-        return (_scale_polygon(self, factor_x, factor_y)
+        return (scale_polygon(self, factor_x, factor_y)
                 if factor_x and factor_y
-                else _scale_contour_degenerate(self._border, factor_x,
-                                               factor_y))
+                else scale_contour_degenerate(self._border, factor_x,
+                                              factor_y))
 
     def translate(self, step_x: Coordinate, step_y: Coordinate) -> 'Polygon':
         """
@@ -825,12 +916,19 @@ class Polygon(Indexable, Shaped):
                     and relation is not Relation.ENCLOSES):
                 raise ValueError('Holes should lie inside border.')
 
-    def _intersect_with_raw_multipolygon(self,
-                                         raw_multipolygon: RawMultipolygon
-                                         ) -> Compound:
-        return from_raw_mix_components(*complete_intersect_multipolygons(
-                [(self._raw_border, self._raw_holes)], raw_multipolygon,
-                accurate=False))
+    def _distance_to_raw_point(self, other: RawPoint) -> Coordinate:
+        return (raw_segment_point_distance(
+                self._path_to_raw_edge(self._raw_point_nearest_path(other)),
+                other)
+                if self._raw_locate(other) is Location.EXTERIOR
+                else 0)
+
+    def _distance_to_raw_segment(self, other: RawSegment) -> Coordinate:
+        other_start, other_end = other
+        return (self._linear_distance_to_raw_segment(other)
+                if (self._raw_locate(other_start) is Location.EXTERIOR
+                    and self._raw_locate(other_end) is Location.EXTERIOR)
+                else 0)
 
     def _intersect_with_raw_multisegment(self,
                                          raw_multisegment: RawMultisegment
@@ -841,6 +939,19 @@ class Polygon(Indexable, Shaped):
                         [(self._raw_border, self._raw_holes)],
                         accurate=False))
 
+    def _linear_distance_to_raw_segment(self, other: RawSegment
+                                        ) -> Coordinate:
+        return raw_segments_distance(
+                self._path_to_raw_edge(self._raw_segment_nearest_path(other)),
+                other)
+
+    def _path_to_raw_edge(self, path: Tuple[int, int]) -> RawSegment:
+        contour_id, edge_index = path
+        raw_contour = (self._holes[contour_id - 1]
+                       if contour_id
+                       else self._border)._raw
+        return raw_contour[edge_index - 1], raw_contour[edge_index]
+
     def _subtract_from_raw_multisegment(self, other_raw: RawMultisegment
                                         ) -> Compound:
         return from_raw_multisegment(subtract_multipolygon_from_multisegment(
@@ -881,9 +992,8 @@ class Polygon(Indexable, Shaped):
                                                          accurate=False))
 
 
-def _polygon_to_centroid_components(polygon: Polygon
-                                    ) -> Tuple[Expansion, Expansion,
-                                               Expansion]:
+def polygon_to_centroid_components(polygon: Polygon
+                                   ) -> Tuple[Expansion, Expansion, Expansion]:
     (x_numerator, y_numerator,
      double_area) = _raw_contour_to_centroid_components(
             polygon._border.to_counterclockwise().raw())
@@ -898,6 +1008,12 @@ def _polygon_to_centroid_components(polygon: Polygon
     return x_numerator, y_numerator, double_area
 
 
+def polygon_to_raw_edges(polygon: Polygon) -> Iterator[RawSegment]:
+    return chain(to_pairs_iterable(polygon._raw_border),
+                 flatten(to_pairs_iterable(raw_hole)
+                         for raw_hole in polygon._raw_holes))
+
+
 def raw_locate_point(raw_polygon: RawPolygon, raw_point: RawPoint) -> Location:
     relation = point_in_polygon(raw_point, raw_polygon)
     return (Location.EXTERIOR
@@ -907,6 +1023,34 @@ def raw_locate_point(raw_polygon: RawPolygon, raw_point: RawPoint) -> Location:
                   else Location.INTERIOR))
 
 
+def scale_polygon(polygon: Polygon,
+                  factor_x: Coordinate,
+                  factor_y: Coordinate) -> Polygon:
+    return Polygon(scale_contour(polygon._border, factor_x, factor_y),
+                   [scale_contour(hole, factor_x, factor_y)
+                    for hole in polygon._holes])
+
+
+def rotate_polygon_around_origin(polygon: Polygon,
+                                 cosine: Coordinate,
+                                 sine: Coordinate) -> Polygon:
+    return Polygon(rotate_contour_around_origin(polygon._border, cosine, sine),
+                   [rotate_contour_around_origin(hole, cosine, sine)
+                    for hole in polygon._holes])
+
+
+def rotate_translate_polygon(polygon: Polygon,
+                             cosine: Coordinate,
+                             sine: Coordinate,
+                             step_x: Coordinate,
+                             step_y: Coordinate) -> Polygon:
+    return Polygon(rotate_translate_contour(polygon._border, cosine, sine,
+                                            step_x, step_y),
+                   [rotate_translate_contour(hole, cosine, sine, step_x,
+                                             step_y)
+                    for hole in polygon._holes])
+
+
 def _raw_contour_to_centroid_components(contour: RawContour
                                         ) -> Tuple[Expansion, Expansion,
                                                    Expansion]:
@@ -924,34 +1068,6 @@ def _raw_contour_to_centroid_components(contour: RawContour
     return x_numerator, y_numerator, double_area
 
 
-def _scale_polygon(polygon: Polygon,
-                   factor_x: Coordinate,
-                   factor_y: Coordinate) -> Polygon:
-    return Polygon(_scale_contour(polygon._border, factor_x, factor_y),
-                   [_scale_contour(hole, factor_x, factor_y)
-                    for hole in polygon._holes])
-
-
-def rotate_polygon_around_origin(polygon: Polygon,
-                                 cosine: Coordinate,
-                                 sine: Coordinate) -> Polygon:
-    return Polygon(rotate_contour_around_origin(polygon._border, cosine, sine),
-                   [rotate_contour_around_origin(hole, cosine, sine)
-                    for hole in polygon._holes])
-
-
-def _rotate_translate_polygon(polygon: Polygon,
-                              cosine: Coordinate,
-                              sine: Coordinate,
-                              step_x: Coordinate,
-                              step_y: Coordinate) -> Polygon:
-    return Polygon(_rotate_translate_contour(polygon._border, cosine, sine,
-                                             step_x, step_y),
-                   [_rotate_translate_contour(hole, cosine, sine, step_x,
-                                              step_y)
-                    for hole in polygon._holes])
-
-
 def _to_endpoints_cross_product_z(start_x: Coordinate,
                                   start_y: Coordinate,
                                   end_x: Coordinate,
@@ -961,3 +1077,94 @@ def _to_endpoints_cross_product_z(start_x: Coordinate,
     return (two_two_diff(minuend, minuend_tail, subtrahend, subtrahend_tail)
             if minuend_tail or subtrahend_tail
             else (minuend - subtrahend,))
+
+
+def _to_raw_point_nearest_path(raw_border: RawContour,
+                               raw_holes: List[RawContour],
+                               raw_point: RawPoint) -> Tuple[int, int]:
+    vertex = raw_border[-1]
+    enumerated_vertices = enumerate(raw_border)
+    min_contour_index = 0
+    min_edge_index, next_vertex = next(enumerated_vertices)
+    squared_distance_to_point = partial(squared_raw_point_segment_distance,
+                                        raw_point)
+    min_squared_distance = squared_distance_to_point((vertex, next_vertex))
+    vertex = next_vertex
+    for edge_index, next_vertex in enumerated_vertices:
+        candidate_squared_distance = squared_distance_to_point((vertex,
+                                                                next_vertex))
+        if candidate_squared_distance < min_squared_distance:
+            min_edge_index, min_squared_distance = (edge_index,
+                                                    candidate_squared_distance)
+        vertex = next_vertex
+    for contour_index, raw_hole in enumerate(raw_holes,
+                                             start=1):
+        vertex = raw_hole[-1]
+        for edge_index, next_vertex in enumerate(raw_hole):
+            candidate_squared_distance = squared_distance_to_point(
+                    (vertex, next_vertex))
+            if candidate_squared_distance < min_squared_distance:
+                (min_contour_index, min_edge_index, min_squared_distance) = (
+                    contour_index, edge_index, candidate_squared_distance)
+            vertex = next_vertex
+    return min_contour_index, min_edge_index
+
+
+def _to_raw_segment_nearest_path(raw_border: RawContour,
+                                 raw_holes: List[RawContour],
+                                 raw_segment: RawSegment) -> Tuple[int, int]:
+    vertex = raw_border[-1]
+    enumerated_vertices = enumerate(raw_border)
+    min_contour_index = 0
+    min_edge_index, next_vertex = next(enumerated_vertices)
+    squared_distance_to_segment = partial(squared_raw_segments_distance,
+                                          raw_segment)
+    min_squared_distance = squared_distance_to_segment((vertex, next_vertex))
+    vertex = next_vertex
+    for index, next_vertex in enumerated_vertices:
+        candidate_squared_distance = squared_distance_to_segment((vertex,
+                                                                  next_vertex))
+        if candidate_squared_distance < min_squared_distance:
+            min_edge_index, min_squared_distance = (index,
+                                                    candidate_squared_distance)
+        vertex = next_vertex
+    for contour_index, raw_hole in enumerate(raw_holes,
+                                             start=1):
+        vertex = raw_hole[-1]
+        for edge_index, next_vertex in enumerate(raw_hole):
+            candidate_squared_distance = squared_distance_to_segment(
+                    (vertex, next_vertex))
+            if candidate_squared_distance < min_squared_distance:
+                (min_contour_index, min_edge_index, min_squared_distance) = (
+                    contour_index, edge_index, candidate_squared_distance)
+            vertex = next_vertex
+    return min_contour_index, min_edge_index
+
+
+def _tree_to_holeless_raw_point_nearest_path(tree: segmental.Tree,
+                                             raw_point: RawPoint
+                                             ) -> Tuple[int, int]:
+    return 0, tree.nearest_to_point_index(raw_point)
+
+
+def _tree_to_holeless_raw_segment_nearest_path(tree: segmental.Tree,
+                                               raw_segment: RawSegment
+                                               ) -> Tuple[int, int]:
+    return 0, tree.nearest_index(raw_segment)
+
+
+def _tree_to_raw_point_nearest_path(tree: segmental.Tree,
+                                    contours_offsets: Tuple[int, ...],
+                                    raw_point: RawPoint) -> Tuple[int, int]:
+    index = tree.nearest_to_point_index(raw_point)
+    contour_index = bisect(contours_offsets, index) - 1
+    return contour_index, index - contours_offsets[contour_index]
+
+
+def _tree_to_raw_segment_nearest_path(tree: segmental.Tree,
+                                      contours_offsets: Tuple[int, ...],
+                                      raw_segment: RawSegment
+                                      ) -> Tuple[int, int]:
+    index = tree.nearest_index(raw_segment)
+    contour_index = bisect(contours_offsets, index) - 1
+    return contour_index, index - contours_offsets[contour_index]
